@@ -1,115 +1,53 @@
-import logging
 import asyncio
-import atexit
-import re
-from typing import Any, Dict, Optional, Callable
+import contextvars
+from typing import Any, Dict, Optional, Callable, Union, Awaitable
+from dataclasses import dataclass
+import logging
+from pathlib import Path
 
-from playwright.sync_api import Locator
-from playwright._impl._page import Page as PageImpl
-from playwright._impl._locator import Locator as LocatorImpl
 from playwright._impl._element_handle import ElementHandle as ElementHandleImpl
+from playwright._impl._locator import Locator as LocatorImpl
+from playwright._impl._page import Page as PageImpl
 from playwright.async_api._generated import (
-    Page as PageAPI,
-    Locator as LocatorAPI,
     ElementHandle as ElementHandleAPI,
 )
+from playwright.async_api._generated import (
+    Locator as LocatorAPI,
+)
+from playwright.async_api._generated import (
+    Page as PageAPI,
+)
+from playwright.sync_api import Route
 from stagehand import StagehandPage
-from .locator import PromptBasedLocator
-from .step import print_pending_modifications, AsyncStepContextManager, execute_step
-from ..agent.agent import Agent
-from ..review import review, ReviewStatus
-from ..constants import REVIEW_TIMEOUT
 
+from .checkpoint import default_checkpoint_callback
+from .locator import PromptBasedLocator
+from .step import AsyncStepContextManager
+from .web_agent import WebAgent
+
+from ..agent.agent import Agent
+from ..constants import REVIEW_TIMEOUT
+from ..review import ReviewStatus, review
+from .dom_utils import (
+    inspect_element_recursive,
+    normalize_locator_string,
+)
+from .warc_utils import parse_warc_file
 
 log = logging.getLogger("va.playwright")
 
+# Context variable to track if we're currently inside a login handler
+_in_login_handler: contextvars.ContextVar[bool] = contextvars.ContextVar(
+    "in_login_handler", default=False
+)
 
-def normalize_locator_string(js_locator: str) -> str:
-    """
-    Normalize a JavaScript-style locator string to Python format.
 
-    Examples:
-    - getByRole('textbox', { name: 'Customer name:' }) -> get_by_role('textbox', name="Customer name:")
-    - getByText('Submit') -> get_by_text('Submit')
-    - getByLabel('Email') -> get_by_label('Email')
-    """
-    if not js_locator:
-        return js_locator
+@dataclass
+class ActResult:
+    """Result of a Page.act operation."""
 
-    # Convert camelCase method names to snake_case
-    # Handle compound words like TestId first
-    js_locator = re.sub(
-        r"getBy([A-Z][a-z]+)([A-Z][a-z]+)",
-        lambda m: f"get_by_{m.group(1).lower()}_{m.group(2).lower()}",
-        js_locator,
-    )
-    # Handle simple cases
-    js_locator = re.sub(
-        r"getBy([A-Z][a-z]*)", lambda m: f"get_by_{m.group(1).lower()}", js_locator
-    )
-
-    # Convert JavaScript object syntax to Python keyword arguments
-    def convert_js_object(match):
-        obj_content = match.group(1).strip()
-
-        # Split by commas, but be careful about commas inside strings
-        parts = []
-        current_part = ""
-        in_string = False
-        quote_char = None
-
-        for char in obj_content:
-            if char in ('"', "'") and not in_string:
-                in_string = True
-                quote_char = char
-                current_part += char
-            elif char == quote_char and in_string:
-                in_string = False
-                quote_char = None
-                current_part += char
-            elif char == "," and not in_string:
-                parts.append(current_part.strip())
-                current_part = ""
-            else:
-                current_part += char
-
-        if current_part.strip():
-            parts.append(current_part.strip())
-
-        # Process each part into keyword arguments
-        processed_parts = []
-        for part in parts:
-            part = part.strip()
-            # Handle key: value pairs
-            if ":" in part:
-                key, value = part.split(":", 1)
-                key = key.strip()
-                value = value.strip()
-
-                # Remove quotes from key if present
-                if key.startswith('"') and key.endswith('"'):
-                    key = key[1:-1]
-                elif key.startswith("'") and key.endswith("'"):
-                    key = key[1:-1]
-
-                # Convert single quotes to double quotes for string values
-                if value.startswith("'") and value.endswith("'"):
-                    value = f'"{value[1:-1]}"'
-                elif value == "true":
-                    value = "True"
-                elif value == "false":
-                    value = "False"
-
-                processed_parts.append(f"{key}={value}")
-            else:
-                processed_parts.append(part)
-
-        return ", ".join(processed_parts)
-
-    # Match JavaScript object literals like { name: 'value', other: 'test' }
-    js_locator = re.sub(r"\{\s*([^}]+)\s*\}", convert_js_object, js_locator)
-
-    return js_locator
+    success: bool
+    message: str
 
 
 # Monkey patch the Playwright Page implementation to add snapshotForAI method
@@ -136,7 +74,7 @@ async def _snapshot_for_ai(self, metadata: Optional[Dict[str, Any]] = None) -> s
 
 
 # Apply the monkey patch to both implementation and API classes
-PageImpl.snapshot_for_ai = _snapshot_for_ai
+PageImpl.snapshot_for_ai = _snapshot_for_ai  # type: ignore
 
 
 # Also add it to the API wrapper class
@@ -145,7 +83,7 @@ async def _api_snapshot_for_ai(self, metadata: Optional[Dict[str, Any]] = None)
     return await self._impl_obj.snapshot_for_ai(metadata)
 
 
-PageAPI.snapshot_for_ai = _api_snapshot_for_ai
+PageAPI.snapshot_for_ai = _api_snapshot_for_ai  # type: ignore
 
 
 # Monkey patch the Playwright ElementHandle implementation to add generateLocatorString method
@@ -168,7 +106,7 @@ async def _element_generate_locator_string(self) -> Optional[str]:
 
 
 # Apply the monkey patch to ElementHandle classes
-ElementHandleImpl._generate_locator_string = _element_generate_locator_string
+ElementHandleImpl._generate_locator_string = _element_generate_locator_string  # type: ignore
 
 
 # Also add it to the API wrapper class
@@ -177,7 +115,7 @@ async def _api_element_generate_locator_string(self) -> Optional[str]:
     return await self._impl_obj._generate_locator_string()
 
 
-ElementHandleAPI._generate_locator_string = _api_element_generate_locator_string
+ElementHandleAPI._generate_locator_string = _api_element_generate_locator_string  # type: ignore
 
 
 # Monkey patch the Playwright Locator implementation to add generateLocatorString method
@@ -200,7 +138,7 @@ async def _generate_locator_string(self) -> Optional[str]:
 
 
 # Apply the monkey patch to both implementation and API classes
-LocatorImpl.generate_locator_string = _generate_locator_string
+LocatorImpl.generate_locator_string = _generate_locator_string  # type: ignore
 
 
 # Also add it to the API wrapper class
@@ -209,11 +147,7 @@ async def _api_generate_locator_string(self) -> Optional[str]:
     return await self._impl_obj.generate_locator_string()
 
 
-LocatorAPI.generate_locator_string = _api_generate_locator_string
-
-
-# Register the cleanup function to run at exit
-atexit.register(print_pending_modifications)
+LocatorAPI.generate_locator_string = _api_generate_locator_string  # type: ignore
 
 
 class Page:
@@ -223,6 +157,11 @@ class Page:
         self._agent = Agent()
         # Track any background login tasks (only for fallback case)
         self._current_login_task = None
+        # WARC serving state
+        self._warc_responses: Optional[Dict[str, Dict[str, Any]]] = None
+        self._warc_route_handler = None
+        # Checkpoint review configuration
+        self._checkpoint_review_callback = None
 
     def get_by_prompt(
         self,
@@ -246,26 +185,6 @@ class Page:
         """
         return PromptBasedLocator(self, prompt)
 
-    async def get_locator_by_prompt(
-        self,
-        prompt: str,
-    ) -> Locator | None:
-        """
-        Internal method to get element by prompt - used by PromptBasedLocator
-
-        Returns:
-        --------
-        Playwright [Locator](https://playwright.dev/python/docs/api/class-locator) | None: The found element or `None` if no matching elements were found.
-        """
-
-        results = await self._stagehand_page.observe(prompt)
-
-        if not results:
-            return None
-
-        selector = results[0].selector
-        return self._stagehand_page.locator(selector)
-
     async def _check_login_and_handle(self, login_handler):
         """Check if login is required and handle it"""
         try:
@@ -273,17 +192,23 @@ class Page:
             result = await self._stagehand_page.extract(
                 "Is login required on this page? Answer with exactly 'yes' or 'no' (lowercase, no additional text)."
             )
-            answer = result.extraction.lower()
+            answer = result.extraction.lower()  # type: ignore
             if answer == "yes":
                 log.info("Login required detected, calling handler")
                 if login_handler:
                     log.info("Calling login handler")
-                    # handle sync and async login handlers
-                    if asyncio.iscoroutinefunction(login_handler):
-                        await login_handler()
-                    else:
-                        login_handler()
-                    log.info("Login handler completed")
+                    # Set context variable to mark we're inside a login handler
+                    _in_login_handler.set(True)
+                    try:
+                        # Always pass the page to the handler
+                        if asyncio.iscoroutinefunction(login_handler):
+                            await login_handler(self)
+                        else:
+                            login_handler(self)
+                        log.info("Login handler completed")
+                    finally:
+                        # Reset context variable when done
+                        _in_login_handler.set(False)
                 else:
                     log.warning("Login required but no handler available")
             else:
@@ -293,6 +218,27 @@ class Page:
             log.error(f"Error in login check: {e}")
             raise Exception(f"Login check failed: {e}") from e
 
+    def enable_checkpoint_review(self, callback=None):
+        """
+        Enable checkpoint review for potentially destructive actions like form submissions and navigation.
+
+        When enabled, the system will pause execution and create a review before performing
+        actions that are likely to submit forms or navigate to different pages.
+
+        Args:
+            callback: Optional custom callback function that takes (command, intent_result, screenshot, code_context)
+                     If not provided, uses default callback that creates a review and waits for completion.
+        """
+        self._checkpoint_review_callback = (
+            callback if callback else default_checkpoint_callback
+        )
+        log.info("Checkpoint review enabled for this page")
+
+    def disable_checkpoint_review(self):
+        """Disable checkpoint review for this page."""
+        self._checkpoint_review_callback = None
+        log.info("Checkpoint review disabled for this page")
+
     def step(
         self,
         command: str,
@@ -305,11 +251,13 @@ class Page:
         This method returns a context manager that can be used with 'async with'.
         The LLM action generation is only triggered if the with block is empty (contains only pass).
 
+        To enable checkpoint reviews for form submissions and navigation, call page.enable_checkpoint_review() first.
+
         Parameters:
         -----------
         command (str): Natural language description of the action to perform
         context (Dict[str, Any], optional): Context variables available to the generated script
-        max_retries (int): Maximum number of retry attempts. Defaults to 3.
+        max_retries (int): Maximum number of retry attempts. Defaults to 16.
 
         Returns:
         --------
@@ -332,24 +280,6 @@ class Page:
         else:
             log.info("Page load detected but no login handler registered")
 
-    async def _execute_step(
-        self,
-        command: str,
-        context: Optional[Dict[str, Any]] = None,
-        max_retries: int = 3,
-    ) -> Dict[str, Any]:
-        """
-        Internal method to execute a natural language command.
-        Now delegates to the pure function in step.py.
-        """
-        return await execute_step(
-            command=command,
-            context=context,
-            max_retries=max_retries,
-            page=self._stagehand_page,
-            agent=self._agent,
-        )
-
     async def _wait_for_login_task(self):
         """Wait for any background login task to complete"""
         if self._current_login_task:
@@ -364,7 +294,13 @@ class Page:
         if not callable(attr):
             return attr
 
-        # TODO: this is a hack to wait for login task to complete before executing the method, this will cause a deadlock if the method has any other page methods. Since HITL won't call any page methods, this is fine for now. We will need to find a better way to handle this.
+        # If we're in a login handler, return the method directly without waiting for login task
+        if _in_login_handler.get(False):
+            return attr
+
+        # Wait for login task to complete before executing methods to prevent actions
+        # from running while login is in progress. The context var prevents deadlocks from
+        # happening if a login handler directly calls page methods.
         if asyncio.iscoroutinefunction(attr):
             # create async wrapper that waits for login task
             async def async_wrapper(*args, **kwargs):
@@ -389,7 +325,13 @@ class Page:
 
             return sync_wrapper
 
-    def on(self, event: str, handler: Optional[Callable[[str], None]] = None):
+    def on(
+        self,
+        event: str,
+        handler: Optional[
+            Union[Callable[[Any], None], Callable[[Any], Awaitable[None]]]
+        ] = None,
+    ):
         """
         Register event handler for page.
 
@@ -399,7 +341,8 @@ class Page:
         Parameters:
         -----------
         event (str): The event to listen for
-        handler (Optional[Callable[[str], None]], optional): The handler function to call when the event occurs.
+        handler (Optional[Union[...]], optional): The handler function to call when the event occurs.
+                   Can be sync or async and will receive the page as a parameter.
                    If not provided for "login_required" event, uses the default login handler.
 
         Examples:
@@ -434,7 +377,11 @@ class Page:
             self._stagehand_page.on(event, handler)
 
     def remove_listener(
-        self, event: str, handler: Optional[Callable[[str], None]] = None
+        self,
+        event: str,
+        handler: Optional[
+            Union[Callable[[Any], None], Callable[[Any], Awaitable[None]]]
+        ] = None,
     ):
         """
         Remove event handler for page. If the event is "login_required", we will remove the listener that checks if login is required on page loads.
@@ -442,7 +389,8 @@ class Page:
         Parameters:
         -----------
         event (str): The event to remove listener for
-        handler (Callable[[str], None] | None): The handler function to remove.
+        handler (Optional[Union[...]], optional): The handler function to remove.
+                   Can be sync or async and will receive the page as a parameter.
         """
         if event == "login_required":
             self._stagehand_page.remove_listener(
@@ -453,9 +401,155 @@ class Page:
         else:
             self._stagehand_page.remove_listener(event, handler)
 
-    def default_login_handler(self):
+    async def default_login_handler(self, page):
         """Default login handler that starts HITL review"""
         r = review("review-for-login", "Please log in to continue.")
-        r.wait(REVIEW_TIMEOUT)  # 1000s timeout
+        await r.wait(REVIEW_TIMEOUT)  # 1000s timeout
         if r.status != ReviewStatus.READY:
             raise Exception("Login review not completed")
+
+    async def act(
+        self, prompt: str, context: Optional[Dict[str, Any]] = None
+    ) -> ActResult:
+        """
+        Execute a natural language action on the page.
+
+        Parameters:
+        -----------
+        prompt (str): Natural language description of the action to perform
+        context (Dict[str, Any], optional): Context variables available to the generated script
+
+        Returns:
+        --------
+        ActResult: Result with success status and message
+        """
+        try:
+            # Create a web agent for this page
+            web_agent = WebAgent(self._stagehand_page)
+
+            # Use provided context or empty dict
+            if context is None:
+                context = {}
+
+            # Execute the action using the web agent
+            result = await web_agent.execute_interactive_step(prompt, context)
+
+            # Return ActResult based on the web agent's result
+            return ActResult(success=result["success"], message=result["message"])
+
+        except Exception as e:
+            log.error(f"Error executing act command '{prompt}': {e}")
+            return ActResult(
+                success=False, message=f"Error executing command: {str(e)}"
+            )
+
+    async def inspect_element(self, x: int, y: int, num_ancestors: int = 3) -> str:
+        """
+        Inspect an element by coordinates.
+
+        This method handles both regular elements and elements within iframes,
+        including nested iframes and cross-origin iframes by using Playwright's frame API.
+
+        Cross-origin iframe behavior:
+        - Uses Playwright's frame API to bypass browser same-origin policy restrictions
+        - Coordinates are automatically translated from main page to iframe coordinate system
+        - Falls back to inspecting the iframe element itself if the frame cannot be accessed
+        - Supports both same-origin and cross-origin iframes seamlessly
+        - Recursively handles nested iframes (iframe inside iframe)
+
+        Args:
+            x: X coordinate of the element
+            y: Y coordinate of the element
+            num_ancestors: Number of ancestors to inspect (default: 3)
+
+        Returns:
+            A condensed path representation of the element and its container
+        """
+        return await inspect_element_recursive(
+            self._stagehand_page, x, y, num_ancestors, mode="element"
+        )
+
+    async def inspect_html(
+        self, x: int, y: int, num_ancestors: int = 3, max_characters: int = 1024
+    ) -> str:
+        """
+        Inspect an element by coordinates and return raw HTML like Chrome DevTools.
+
+        This method handles both regular elements and elements within iframes,
+        including nested iframes and cross-origin iframes by using Playwright's frame API.
+
+        Cross-origin iframe behavior:
+        - Uses Playwright's frame API to bypass browser same-origin policy restrictions
+        - Coordinates are automatically translated from main page to iframe coordinate system
+        - Falls back to inspecting the iframe element itself if the frame cannot be accessed
+        - Supports both same-origin and cross-origin iframes seamlessly
+        - Recursively handles nested iframes (iframe inside iframe)
+
+        Args:
+            x: X coordinate of the element
+            y: Y coordinate of the element
+            num_ancestors: Number of ancestors to inspect (default: 3)
+            max_characters: Maximum characters in response.
+
+        Returns:
+            Raw HTML content of the element and its container, formatted like Chrome DevTools
+        """
+        return await inspect_element_recursive(
+            self._stagehand_page,
+            x,
+            y,
+            num_ancestors,
+            mode="html",
+            max_characters=max_characters,
+        )
+
+    async def serve_from_warc(
+        self, warc_file_path: Path, fallback_to_live: bool = False
+    ):
+        """
+        Configure the page to serve responses from a WARC file.
+        Args:
+            warc_file_path: Path to the WARC file
+            fallback_to_live: If True, fall back to live requests for URLs not in WARC.
+                             If False, return 404 for missing URLs.
+        """
+        # Parse the WARC file
+        self._warc_responses = parse_warc_file(warc_file_path)
+
+        # Set up request interception
+        async def warc_route_handler(route: Route):
+            request_url = route.request.url
+
+            # Check if URL exists in WARC
+            if self._warc_responses and request_url in self._warc_responses:
+                warc_response = self._warc_responses[request_url]
+                await route.fulfill(
+                    status=warc_response["status"],
+                    headers=warc_response["headers"],
+                    body=warc_response["body"],
+                )
+            elif fallback_to_live:
+                # Allow live request
+                await route.continue_()
+            else:
+                # Return 404 for URLs not in WARC
+                await route.fulfill(
+                    status=404,
+                    body=f"URL not found in WARC: {request_url}",
+                )
+
+        # Store handler reference for cleanup
+        self._warc_route_handler = warc_route_handler
+
+        # Register the route handler
+        await self.route("**/*", warc_route_handler)
+
+    async def clear_warc_serving(self):
+        """Clear WARC serving configuration and remove route handler."""
+        if self._warc_route_handler:
+            await self.unroute("**/*", self._warc_route_handler)
+            self._warc_route_handler = None
+            self._warc_responses = None
+
+    async def get_page_snapshot(self):
+        return self.snapshot_for_ai()
