@@ -1,65 +1,63 @@
-import base64
 import json
 import logging
-from typing import Any, Dict, List, Optional, Tuple
+from typing import Any, Dict, List, Optional
 from dataclasses import dataclass
 
 from playwright.async_api import Page
 from ..agent.agent import Agent
+from .tools import PLAYWRIGHT_TOOLS, GENERIC_TOOLS, ALL_TOOL_HANDLERS
+from mcp import types
 
 
 log = logging.getLogger("va.playwright.web_agent")
 
 
-# Define available tools (Anthropic format)
-TOOLS = [
-    {
-        "name": "get_page_snapshot",
-        "description": "Get an AI-optimized snapshot of the current page structure",
-        "input_schema": {
-            "type": "object",
-            "properties": {},
-            "required": [],
-        },
-    },
-    {
-        "name": "execute_python_command",
-        "description": "Execute a Python command to interact with the page",
-        "input_schema": {
-            "type": "object",
-            "properties": {
-                "command": {
-                    "type": "string",
-                    "description": "Python command to execute",
+# Utility functions to convert between MCP types and Anthropic format
+def mcp_tool_to_anthropic_tool(mcp_tool: types.Tool) -> Dict[str, Any]:
+    """Convert MCP Tool to Anthropic tool format."""
+    return {
+        "name": mcp_tool.name,
+        "description": mcp_tool.description,
+        "input_schema": mcp_tool.inputSchema,
+    }
+
+
+def mcp_content_to_anthropic_content(
+    mcp_content: List[types.TextContent | types.ImageContent],
+) -> Any:
+    """Convert MCP content to Anthropic format."""
+    if len(mcp_content) == 1:
+        content = mcp_content[0]
+        if isinstance(content, types.TextContent):
+            return content.text
+        elif isinstance(content, types.ImageContent):
+            return [
+                {
+                    "type": "image",
+                    "source": {
+                        "type": "base64",
+                        "media_type": content.mimeType,
+                        "data": content.data,
+                    },
                 }
+            ]
+    return [
+        content.text
+        if isinstance(content, types.TextContent)
+        else {
+            "type": "image",
+            "source": {
+                "type": "base64",
+                "media_type": content.mimeType,
+                "data": content.data,
             },
-            "required": ["command"],
-        },
-    },
-    {
-        "name": "find_element_by_ref",
-        "description": "Find an element by its ref from the AI snapshot",
-        "input_schema": {
-            "type": "object",
-            "properties": {
-                "ref": {
-                    "type": "string",
-                    "description": "Element ref (e.g., 'e3')",
-                }
-            },
-            "required": ["ref"],
-        },
-    },
-    {
-        "name": "get_page_screenshot",
-        "description": "Take a screenshot of the current page",
-        "input_schema": {
-            "type": "object",
-            "properties": {},
-            "required": [],
-        },
-    },
-]
+        }
+        for content in mcp_content
+    ]
+
+
+# Convert all tools to Anthropic format
+TOOLS = [mcp_tool_to_anthropic_tool(tool) for tool in PLAYWRIGHT_TOOLS + GENERIC_TOOLS]
 
 
 def build_system_prompt(command: str, context: Dict[str, Any]) -> str:
@@ -188,88 +186,6 @@ class WebAgent(Agent):
                     result_content = result_content[:197] + "..."
                 log_fn(f"  {j}. {result_content}")
 
-    async def get_page_snapshot(self) -> str:
-        """Tool: Get an AI-optimized snapshot of the current page."""
-        try:
-            # Use our monkey-patched method
-            snapshot = await self.page.snapshot_for_ai()
-            return snapshot
-        except Exception as e:
-            log.error(f"Failed to get page snapshot: {e}")
-            return f"Error getting page snapshot: {e}"
-
-    async def execute_python_command(
-        self, command: str, context: Dict[str, Any]
-    ) -> Tuple[bool, str]:
-        """Tool: Execute a Python command on the page and return success status and result."""
-        try:
-            # Create execution context with captured output
-            captured_output = []
-            execution_context = {
-                "page": self.page,
-                "print": lambda *args: captured_output.append(
-                    " ".join(str(arg) for arg in args)
-                ),
-                "context": context,
-            }
-
-            result = None
-            # Handle async commands
-            if "await " in command:
-                # Wrap in async function and capture return value
-                async_command = f"""
-async def __execute_command():
-    {command.replace(chr(10), chr(10) + "    ")}
-    
-__result = __execute_command()
-"""
-                exec(compile(async_command, "<string>", "exec"), execution_context)
-                if "__result" in execution_context:
-                    result = await execution_context["__result"]
-            else:
-                # Execute synchronous command and capture return value
-                result = exec(command, execution_context)
-
-            # Build response message
-            response_parts = []
-            if captured_output:
-                response_parts.append("Output: " + "\n".join(captured_output))
-            if result is not None:
-                response_parts.append(f"Return value: {result}")
-            if not response_parts:
-                response_parts.append("Command executed successfully")
-
-            return True, "\n".join(response_parts)
-        except Exception as e:
-            return False, f"Error executing command: {e}"
-
-    async def find_element_by_ref(self, ref: str) -> Optional[str]:
-        """Tool: Find an element by its ref from the AI snapshot and return a locator string."""
-        try:
-            locator = self.page.locator(f"aria-ref={ref}")
-            count = await locator.count()
-
-            if count > 0:
-                # Generate locator string for the element
-                locator_string = await locator.first.generate_locator_string()
-                return "page." + locator_string
-            else:
-                return None
-        except Exception as e:
-            log.error(f"Failed to find element by ref {ref}: {e}")
-            return None
-
-    async def get_page_screenshot(self) -> str:
-        """Tool: Take a screenshot of the current page."""
-        try:
-            # Take screenshot and return as base64 string
-            screenshot_bytes = await self.page.screenshot()
-            screenshot_base64 = base64.b64encode(screenshot_bytes).decode("utf-8")
-            return screenshot_base64
-        except Exception as e:
-            log.error(f"Failed to take screenshot: {e}")
-            return f"Error taking screenshot: {e}"
-
     async def process_anthropic_tool_call(
         self, tool_call, context: Dict[str, Any]
     ) -> Dict[str, Any]:
@@ -278,59 +194,36 @@ __result = __execute_command()
         arguments = tool_call.input
 
         try:
-            if tool_name == "get_page_snapshot":
-                result = await self.get_page_snapshot()
-                return {
-                    "type": "tool_result",
-                    "tool_use_id": tool_call.id,
-                    "content": result,
-                }
+            if tool_name in ALL_TOOL_HANDLERS:
+                # Use centralized tool handlers
+                handler = ALL_TOOL_HANDLERS[tool_name]
 
-            elif tool_name == "execute_python_command":
-                command = arguments.get("command", "")
-                success, result = await self.execute_python_command(command, context)
-                # Add successful commands to our working script
-                if success and not command.strip().startswith("print"):
-                    self.working_script_lines.append(command)
-                return {
-                    "type": "tool_result",
-                    "tool_use_id": tool_call.id,
-                    "content": f"Success: {success}\n{result}",
-                }
+                # For execute_python_command, add context support and handle working script
+                if tool_name == "execute_python_command":
+                    # Use centralized handler with context support
+                    mcp_content = await handler(self.page, arguments, context)
+                    anthropic_content = mcp_content_to_anthropic_content(mcp_content)
 
-            elif tool_name == "find_element_by_ref":
-                ref = arguments.get("ref", "")
-                result = await self.find_element_by_ref(ref)
-                return {
-                    "type": "tool_result",
-                    "tool_use_id": tool_call.id,
-                    "content": result or "Element not found",
-                }
+                    # Add successful commands to our working script
+                    command = arguments.get("command", "")
+                    if not command.strip().startswith(
+                        "print"
+                    ) and "Success: True" in str(anthropic_content):
+                        self.working_script_lines.append(command)
 
-            elif tool_name == "get_page_screenshot":
-                result = await self.get_page_screenshot()
-                if result.startswith("Error"):
-                    # Return error as text
                     return {
                         "type": "tool_result",
                         "tool_use_id": tool_call.id,
-                        "content": result,
+                        "content": anthropic_content,
                     }
                 else:
-                    # Return image in proper format
+                    # Use the centralized handler for other tools
+                    mcp_content = await handler(self.page, arguments)
+                    anthropic_content = mcp_content_to_anthropic_content(mcp_content)
                     return {
                         "type": "tool_result",
                         "tool_use_id": tool_call.id,
-                        "content": [
-                            {
-                                "type": "image",
-                                "source": {
-                                    "type": "base64",
-                                    "media_type": "image/png",
-                                    "data": result,
-                                },
-                            }
-                        ],
+                        "content": anthropic_content,
                     }
 
             else:
